/*-
# X-BASED ABACUS
#
#  AbacusC.c
#
###
#
#  Copyright (c) 1992 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "useful",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  Taken from a C++ group project where I was a lead contributer
#  OOP Group4
*/

/* Methods string calculator file for Abacus */

#if WINVER
#include "wabacus.h"
#else
#include <stdlib.h>
#include <stdio.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "AbacusP.h"
#endif

#ifdef CALC
enum OrderType {notused, low, medium, high, unary, mode, constant, equate};
static const short unsigned int orderASCII[] =
{
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,
   notused,  notused,  notused,  notused,

   notused,    unary,  notused, constant, /* Space ! " # */
  constant,   medium,   medium,  notused, /* $ % & ' */
    equate,   equate,   medium,      low, /* ( ) * + */
   notused,      low,  notused,   medium, /* , - . / */
   notused,  notused,  notused,  notused, /* 0 1 2 3 */
   notused,  notused,  notused,  notused, /* 4 5 6 7 */
   notused,  notused,  notused,  notused, /* 8 9 : ; */
      high,   equate,     high,    unary, /* < = > ? */


  constant,  notused,  notused,  notused, /* @ A B C */
   notused,  notused,  notused, constant, /* D E F G */
      mode,     mode,  notused,  notused, /* H I J K */
     unary, constant,    unary,     high, /* L M N O */
  constant,  notused,     high, constant, /* P Q R S */
  constant,    unary,    unary,    unary, /* T U V W */
     unary,  notused,  notused,  notused, /* X Y Z [ */
    medium,  notused,     high,     mode, /* \ ] ^ _ */

   notused,  notused,     mode,      low, /* ` a b c */
      mode, constant,  notused,     mode, /* d e f g */
      mode,    unary,  notused,    unary, /* h i j k */
     unary,     mode,    unary,     mode, /* l m n o */
       low,  notused,     mode,    unary, /* p q r s */
     unary,    unary,    unary,    unary, /* t u v w */
      mode,  notused,  notused,  notused, /* x y z { */
       low,  notused,    unary,  notused  /* | } ~ Delete */
};

enum OperationType {undefined, ignore, numeric, operation, clear, quit};
static const short unsigned int operationASCII[] =
{
  undefined, undefined, undefined,      quit, /* Null SOH STX Interrupt */
  undefined, undefined, undefined, undefined, /* EOT ENQ ACK Bell */
    numeric, undefined,    ignore, undefined, /* Backspace Tab Linefeed VT */
  undefined,    ignore, undefined, undefined, /* Formfeed Carriagereturn SO SI */
  undefined, undefined, undefined, undefined, /* DLE DC1 DC2 DC3 */
  undefined, undefined, undefined, undefined, /* DC4 NAK SYN ETB */
  undefined, undefined, undefined, undefined, /* CAN EM SUB Escape */
  undefined, undefined, undefined, undefined, /* FS GS RS US */

     ignore, operation, undefined, operation, /* Space ! " # */
  operation, operation, operation, undefined, /* $ % & ' */
  operation, operation, operation, operation, /* ( ) * + */
    numeric, operation,   numeric, operation, /* , - . / */
    numeric,   numeric,   numeric,   numeric, /* 0 1 2 3 */
    numeric,   numeric,   numeric,   numeric, /* 4 5 6 7 */
    numeric,   numeric, undefined, undefined, /* 8 9 : ; */
  operation, operation, operation, operation, /* < = > ? */

  operation,   numeric,   numeric,   numeric, /* @ A B C */
    numeric,   numeric,   numeric, operation, /* D E F G */
  operation, operation, undefined, undefined, /* H I J K */
  operation, operation, operation, operation, /* L M N O */
  operation,      quit, operation, operation, /* P Q R S */
  operation, operation, operation, operation, /* T U V W */
  operation, undefined, undefined, undefined, /* X Y Z [ */
  operation, undefined, operation, operation, /* \ ] ^ _ */

  undefined,     clear, operation, operation, /* ` a b c */
  operation, operation, undefined, operation, /* d e f g */
  operation, operation, undefined, operation, /* h i j k */
  operation, operation, operation, operation, /* l m n o */
  operation,     clear, operation, operation, /* p q r s */
  operation, operation, operation, operation, /* t u v w */
  operation, undefined,     clear, undefined, /* x y z { */
  operation, undefined,   numeric, undefined  /* | } ~ Delete */
};

#define CHAR_TO_DIGIT(c) ((c >= 'A') ? c - 'A' + 10 : c - '0')
#define DIGIT_TO_CHAR(d) ((d >= 10) ? 'A' + d - 10 : '0' + d)
#define IS_DIGIT(c) ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F'))
/* True for period, +/-, or operator */
#define IS_VALID(v, b) ((IS_DIGIT(v)) ? (CHAR_TO_DIGIT(v) < b) : True)
#define MAXVALUELENGTH 64

enum StackUsage {paren, order};

typedef struct _Term
{
	double variable;
	char operation;
	Boolean orderUsage;
	struct _Term *previous;
} Term;

typedef struct _Expression
{
	/* char *variable; */
	char variable[MAXVALUELENGTH]; /* Hard code a limit for now */
	char operation;
	struct _Expression *previous;
} Expression;

/* struct Stack */
static Term *term;
static Expression *expression;

#define DEFAULTBASE 10
#define DEFAULTFRACTDIGITS 8

/* Inputter & Parser & Paren Manager */
static Boolean period = False;
static Boolean gotFirstDigit = False; /* handles special case of just a '0' */
static double left, right; /* sides of a binary operation */
static char pendingOperation; /* operation to be performed on left (& right if not unary) */
static Boolean hub; /* intermediate result? usually a ')' or pi pressed */
static char displayBuf[64];
static int digits = 1;
static int currentBase = DEFAULTBASE, fractDigits = DEFAULTFRACTDIGITS;
static int nestingLevel = 0;
static Expression parseExpression;


#if 0
static double
cbrt(double x) /* -inf < x < inf */
{
	return ((x < 0.0) ? -pow(-x, 1.0 / 3.0) : pow(x, 1.0 / 3.0));
}
#endif

static double
powInt(int base, int y)
{
	int i;
	double z = 1.0;

	if (y > 0)
		for (i = 0; i < y; i++)
			z *= base;
	else if (y < 0)
		for (i = 0; i > y; i--)
			z /= base;
	else
		z = 1.0;
	return z;
}

/* More exact (int) (log(x) / log ((double) base)) */
static int
logInt(double x, int base)
{
	int i = 0;

	if (x >= (double) base)
		while (x >= (double) base) {
			x /= (double) base;
			i++;
		}
	else if (x > 0.0)
		while (x < 1.0) {
			x *= (double) base;
			i--;
		}
	return i;
}

static double
rootInt(double x, int y) /* y != 0 && (x >= 0 || (y odd)) */
{
	if (x < 0.0 && ((double) y / 2.0) != ((double) (y / 2)))
		return -pow (-x, 1.0 / (double) y);
	else if (x == 0.0)
		return 0.0;
	else
		return pow (x, 1.0 / (double) y);
}

static double
convertToDecimal(int b, char *inputstring)
{
	int k = 0;
	Boolean negative = False;
	double number = 0.0;
	int digit;
	int length = 0;
	double factor;

	printf("DecimalIn: %s\n", inputstring);
	/* Convert Integer Part */
	k = 0;
	if (inputstring[k] == '-')
	{
		negative = True;
		k++;
	}
	while (IS_DIGIT(inputstring[k + length]))
		length++;
	factor = powInt(b, length);
	for (; length > 0; length--, k++)
	{
		digit = CHAR_TO_DIGIT(inputstring[k]);
		factor /= ((double) b);
		number += ((double) digit) * factor;
	}

	/* Convert Fractional Part */
	if (inputstring[k] == '.')
	{
		k++;
		while (IS_DIGIT(inputstring[k]))
		{
			digit = CHAR_TO_DIGIT(inputstring[k]);
			factor /= ((double) b);
			number += ((double) digit) * factor;
			k++;
		}
	}

	if ((Boolean) negative)
		number = -number;
	negative = False;

(void) printf("DecimalOut: %g\n", number);
	return number;
}

static void
convertFromDecimal(char *outputstring, int b, double x)
{
	char string[MAXVALUELENGTH];
	Boolean localPeriod = False;
	double number = 0.0;
	int places_base;
	int l = 0, i= 0;
	short unsigned int digit;
	double factor;
	int fractdigits = fractDigits;

	/* (void) printf("\nDecimalIn: %g\n", x); */
	(void) sprintf(string, "%g", x);
	if (string[i] == '-') {
		outputstring[l++] = '-';
		x = -x;
	}
	while (string[i] != '\0') {
		if (string[i] == '.')
			localPeriod = True;
		i++;
	}
	{
		/* Chicken and egg problem: rounding might increase places_base */
		places_base = logInt(x, b);
		fractdigits -= (places_base + 1);
		x += (pow(b, -fractdigits) / 2.0); /* rounder */

		number = x;
		/* Convert Integer Part */
		if (number < 1.0) {
			outputstring[l++] = '0';
			factor = 1.0 / ((double) b);
		} else {
			places_base = logInt(number, b);
			factor = pow(b, places_base);
			for (; places_base >= 0; places_base--) {
				digit = (int) (number / factor);
				outputstring[l++] = DIGIT_TO_CHAR(digit);
				number -= digit * factor;
				factor /= ((double) b);
			}
		}
		/* Convert Fractional Part */
		if (localPeriod) {
(void) printf("\nHI%g %g %gHI\n", number, factor, (double) b);
			outputstring[l++] = '.';
			for (places_base = 1; places_base <= fractdigits; places_base++) {
				digit = (int) (number / factor);
				outputstring[l++] = DIGIT_TO_CHAR(digit);
				number -= digit * factor;
				factor /= ((double) b);
			}
			while (outputstring[l - 1] == '0')
				l--;
		}
	}

	outputstring[l] = '\0';
(void) printf("\nHI%sHI\n", outputstring);
}

static double
formatFromDisplay(int base, char *string)
{
	char floatString[MAXVALUELENGTH];
	Boolean got1Digit = False;
	short unsigned int periods = 0;
	int s = 0, k = 0;

(void) printf("preDecimal: %s\n", string);
	while (string[s] != '\0') { /* Look at each input character */
		if (IS_DIGIT(string[s])) {
			got1Digit = True;
			floatString[k++] = string[s];
		} else switch (string[s]) {
		case '-':
			floatString[k++] = '-';
			break;
		case '.':
			if (periods == 1)
				(void) printf("'%c' handler not implemented in Format_From_Display\n", string[s]);
			periods++;
			if (!got1Digit) {
				got1Digit = True;
				floatString[k++] ='0';
			}
			floatString[k++] = '.';
			break;
		default:
			(void) printf("'%c' handler not implemented in Format_From_Display\n", string[s]);
		}
		s++;
	}
	if (!got1Digit) {
		got1Digit = True;
		floatString[k++] = '0';
	}
	floatString[k] = '\0';
	return convertToDecimal(base, floatString);
}

static void
formatToDisplay (char *string, double z)
{
	int i = 0, j, length;
	short unsigned int periods = 0;

	convertFromDecimal(string, currentBase, z);
	length = strlen(string);
	while (i <= length) {
		if (string[i] == '\0' && periods == 0) {
			j = length;
			while (j >= i)
			{
				string[j + 1] = string[j];
				j--;
			}
			length++;
			string[i++] = '.';
			periods++;
		}
		if (string[i] == '.')
			periods++;
		if (string[i] == '+') {
			j = i;
			while (j < length) {
				string[j] = string[j + 1];
				j++;
			}
			length--;
		}
		else
			i++;
	}
	string[length] = '\0';
}


static void reset(void)
{
	period = False;
	gotFirstDigit = False;
	displayBuf[0] = '0';
	displayBuf[1] = '\0';
	digits = 1;
}

static void setBase(int base)
{
	currentBase = base;
	fractDigits = logInt(powInt(DEFAULTBASE, DEFAULTFRACTDIGITS), base);
}

static int getNestingLevel(void)
{
	return nestingLevel;
}

static void incNestingLevel(void)
{
	nestingLevel++;
}

static void decNestingLevel(void)
{
	nestingLevel--;
	if (nestingLevel < 0) /* Ignore extra right parentheses */
		nestingLevel = 0;
}

static void resetNestingLevel(void)
{
	nestingLevel = 0;
}

static double
evaluateSingle(double arg, char operation)
{
	double comp = 0.0;

	switch(operation) {
	case 'i': /* (1/0) undefined and distint from (n/0), where n in R */
		comp = 1.0 / arg;
		break;
	case '!':
		if (comp <= 0.0)
			return 0.0;
		comp = arg; /* TODO factorial */
		break;
	case 'v':
		if (comp <= 0.0)
			return 0.0;
		comp = sqrt(arg);
		break;
	case 'u':
		comp = cbrt(arg);
		break;
	default:
		return 0.0;
	}
	return comp;
}

static double
evaluateDouble(double arg1, char operation, double arg2)
{
	double comp = 0.0;

	switch(operation) {
	case '+':
		comp = arg1 + arg2;
		break;
	case '-':
		comp = arg1 - arg2;
		break;
	case '*':
		comp = arg1 * arg2;
		break;
	case '/':
		if (comp == 0.0)
			return 0.0;
		comp = arg1 / arg2;
		break;
	case '^':
		comp = pow(arg1, arg2);
		break;
	default:
		return 0.0;
	}
	return comp;
}

static Boolean
initStack(void)
{
	Term *plate;

	if (!(plate = (Term *) malloc(sizeof(Term))))
		return False;
	plate->previous = NULL;
	term = plate->previous;
	return True;
}

static Boolean
emptyStack(void)
{
	return ((Boolean) (term == NULL));
}

static void
flushStack(void)
{
	while (!emptyStack()) {
		Term *plate = term;

		term = plate->previous;
		free(plate);
	}
}

static void
delStack(void)
{
	flushStack();
	/* free(term); */
}

static Boolean
pushStack(double z, char c, Boolean u)
{
	Term *plate;

	if (!(plate = (Term *) malloc(sizeof(Term))))
		return False;
	plate->variable = z;
	plate->operation = c;
	plate->orderUsage = u;
	plate->previous = term;
	term = plate;
	return True;
}

static void popStack(double *z, char *c, Boolean *u)
{
	Term *plate = term;

	*z = term->variable;
	*c = term->operation;
	*u = term->orderUsage;
	term = plate->previous;
	free(plate);
}

static char topOp(void)
{
	return term->operation;
}

static Boolean topUsage(void)
{
	return term->orderUsage;
}

static Boolean emptyExpression(void)
{
	return (emptyStack());
}

static Boolean previousOrder(void)
{
	return ((Boolean) (!emptyExpression() && topUsage() == order));
}

static Boolean canReduceExpression(char operation)
{
	return ((Boolean) (previousOrder() &&
		orderASCII[(int) topOp()] >= orderASCII[(int) operation]));
}

static void resetWholeExpression(void)
{
	flushStack();
}

static void getPreviousExpressionPart(double *myLeft, char *operation)
{
	Boolean dummy;

	popStack(myLeft, operation, &dummy);
}

static Boolean storeExpressionParen(double *myLeft, char *operation)
{
	if (pushStack(*myLeft, *operation, paren)) {
		*myLeft = 0.0;
		*operation = '\0';
		return True;
	}
	return False;
}

static void evaluateExpressionPart(double *myLeft, char operation,
		double myRight)
{
	if (operation != '\0')
		*myLeft = evaluateDouble(*myLeft, operation, myRight);
	else
		*myLeft = myRight;
}

static Boolean evaluateExpressionOrder(double *myLeft, char *operation,
	double *myRight, char newOperation)
{
	if (*operation != '\0') {
		if (orderASCII[(int) *operation] >= orderASCII[(int) newOperation]) {
			*myLeft = evaluateDouble(*myLeft, *operation, *myRight);
			while (canReduceExpression(newOperation)) {
				*myRight = *myLeft;
				getPreviousExpressionPart (myLeft, operation);
				evaluateExpressionPart(myLeft, *operation, *myRight);
			}
		} else {
			if (pushStack(*myLeft, *operation, order)) {
				*myLeft = *myRight;
				return True;
			} else
				return False;
		}
	} else
		*myLeft = *myRight;
	return True;
}

static void evaluateExpressionParen(double *myLeft, char *operation)
{
	double myRight;

	while (!emptyExpression() && topUsage() == order) {
		myRight = *myLeft;
		getPreviousExpressionPart(myLeft, operation);
		evaluateExpressionPart(myLeft, *operation, myRight);
	}
}

static void evaluateExpression(double *myLeft, char *operation)
{
	double myRight;

	while (!emptyExpression()) {
		myRight = *myLeft;
		getPreviousExpressionPart(myLeft, operation);
		if (operation != NULL)
			*myLeft = evaluateDouble(*myLeft, *operation, myRight);
		else
			*myLeft = myRight;
	} /* ignore uneven "(()" */
	*operation = '\0';
}

static void
initParser(void)
{
	left = 0.0;
	right = 0.0;
	pendingOperation = '\0';
	hub = False;
	/* FIXME expression = Expression(); */
}

static void
resetRight(void)
{
	right = 0.0;
	hub = True;
}

static void
resetExpression(void)
{
	right = left = 0.0;
	pendingOperation = '\0';
	hub = False;
	resetWholeExpression();
}

static void
parse(double input, Boolean numeral, char newOperation)
{
(void) printf("got here %g %d %c\n", input, numeral, newOperation);
	if (numeral)
		right = input;
	else if (hub) {
		numeral = True;
		hub = False;
	}
	if (orderASCII[(int) newOperation] == equate) {
		switch (newOperation) {
		case '(':
			incNestingLevel();
			storeExpressionParen(&left, &pendingOperation);
			right = left; /* "#(" # is lost forever */
			break;
		case ')': /* Evaluate last term and restore variable and operation. */
			decNestingLevel();
			if (pendingOperation != '\0' && numeral) /* this stops 4+) = 8 */
				left = evaluateDouble(left, pendingOperation, right);
			if (pendingOperation == '\0' && numeral)
				left = right;
			else {
				evaluateExpressionParen(&left, &pendingOperation);
				pendingOperation = '\0';
				right = left;
				if (!emptyExpression()) /* ignore uneven "())" */
				{
					getPreviousExpressionPart (&left, &pendingOperation);
					hub = True;
				}
			}
			break;
		case '=': /* Evaluate term and restore variable and operation. */
(void) printf("even got here %g %d %c\n", input, numeral, newOperation);
			resetNestingLevel();
			if (pendingOperation != '\0' && numeral) /* this stops 4+= = 8 (2+2) */
				left = evaluateDouble(left, pendingOperation, right);
			if (pendingOperation == '\0' && numeral)
				left = right;
			else
			{
				evaluateExpression(&left, &pendingOperation);
				right = left;
			}
			break;
		default:
			resetExpression();
		}
	} else if (orderASCII[(int) newOperation] == constant) {
		switch (newOperation) {
		case 'e':	/* natural logarithm */
			right = M_E;
			break;
		case 'P':	/* pi */
			right = M_PI;
			break;
		default:
			resetExpression();
		}
		hub = True;
	} else if (orderASCII[(int) newOperation] == unary) {
		right = evaluateSingle(right, newOperation);
		if (pendingOperation == '\0')
			left = right;
		hub = True;
	} else if (orderASCII[(int) newOperation] != notused) /* binary operation */ {
(void) printf("may have got here %g %d %c %g %g\n", input, numeral, newOperation, left, right);
		if (!numeral) {
			if (previousOrder()) /* this stops 2*+ = 6 (2*2+2) */ {
				getPreviousExpressionPart(&left, &pendingOperation);
				evaluateExpressionOrder(&left, &pendingOperation, &right,
					newOperation);
				right = left;
			} else if (pendingOperation != '\0')
				resetExpression();
		} else {
			evaluateExpressionOrder(&left, &pendingOperation, &right,
				newOperation);
			right = left;
		}
		pendingOperation = newOperation;
	} else /* not used operation */
		resetExpression();
(void) printf("answer %g\n", right);
	/* TODO */
	formatToDisplay(displayBuf, right);
(void) printf("real answer %s\n", displayBuf);

	/* updateDisplay(displayBuf); */
}

static Boolean
inputOperator(char input)
{
	if (!gotFirstDigit)
		parse(0.0, False, input);
	else {
		parse(formatFromDisplay(currentBase, displayBuf), True, input);
	}
	reset();
	return True;
}

static Boolean
inputNumeric(char input)
{
	int i = digits - 1, j;

(void) printf("displayBuf: %c concatenate with %s\n", input, displayBuf);
	if (!IS_VALID(input, currentBase)) {
		reset();
		return False;
	}
	if (IS_DIGIT(input)) {
		gotFirstDigit = True;
		if (displayBuf[0] == '0' && displayBuf[1] == '\0')
			displayBuf[0] = input;
		else
			displayBuf[digits++] = input;
	} else switch(input) {
	case '.': /* Handle multiple '.' */
		if (period) {
			reset();
			return False;
		}
		period = True;
		displayBuf[digits++] = input;
		break;
	case '~': /* Handle +/- */
	/* Note: if `displayBuf == "0"' then "+/-" should act as a unary operation,
	that is, if no number is entered, it should act on the result. */
		if (!gotFirstDigit) {
			inputOperator ('~');
			return False;
		}
		if (displayBuf[0] == '0' && displayBuf[1] == '\0')
			return False;
		while (displayBuf[i] != ' ' && displayBuf[i] != '-' &&
			displayBuf[i] != ',' && i != 0)
			i--;
		if (displayBuf[i] == '-') {
			/* Take a '-' from front of a negative number */
			while (displayBuf[i] != '\0') {
				displayBuf[i] = displayBuf[i + 1];
				i++;
			}
			digits--;
		} else {
			/* Put the '-' in the front of a positive number */
			if (digits == MAXVALUELENGTH - 1)
				return False;
			j = digits;
			while (j >= i) {
				displayBuf[j + 1] = displayBuf[j];
				j--;
			}
			displayBuf[i] = '-';
			digits++;
		}
		break;
	default:
		reset();
		return False;
	}
(void) printf("displayBuf: %s\n", displayBuf);
	return True;
}


void doTheMath(AbacusWidget w, char * buffer)
{
	unsigned int i;

(void) printf("dothedo #%s#\n", buffer);
(void) printf("dothedo #%s#\n", strpbrk(buffer, "+-*/^!()="));
	resetExpression();
	displayBuf[0] = '0';
	displayBuf[1] = '\0';
	for (i = 0; i < strlen(buffer); i++) {
		switch (operationASCII[(int) buffer[i]]) {
		case undefined:
			(void) printf("undefined\n");
			break;
		case ignore:
			(void) printf("ignore\n");
			break;
		case numeric:
			(void) printf("number %c\n", buffer[i]);
			inputNumeric(buffer[i]);
			break;
		case operation:
			(void) printf("operate %c\n", buffer[i]);
			inputOperator(buffer[i]);
			break;
		case clear:
			(void) printf("clear\n");
			break;
		default:
			(void) printf("QUIT\n");
			break;
		}
	}
(void) printf("the real answer %s\n", displayBuf);
	SetAbacusMove(w, ABACUS_MOVE, 0, 1, 1, 1);
}
#endif
